/*-------------------------------------------------------------------*/
/* CBSOM.C        Pasi Frnti                                        */
/*                                                                   */
/* Kohonen self-organizing feature map (SOM) for VQ                  */
/* Iterates: I*N times in total (I=Iterations, N=Neigh.Size)         */
/*                                                                   */
/* Supported new optional features:                                  */
/*     - Iterations (I) inversely proportional to Neigh.Size         */
/*	   - Initialization via randomly chosen training vectors		 */
/*     - Search for nearest neuron limiting to the neighborhood      */
/*     - Stochastic weighting for neuron updates                     */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName	   "CBSOM"
#define VersionNumber  "Version 0.03"
#define LastUpdated    "6.12.98"

/* ----------------------------------------------------------------- */

#include <stdio.h>
#include <math.h>
#include <time.h>
#include <values.h>

#define  FACTFILE  "cbsom.fac"

#include "parametr.c"
#include "cb.h"
#include "file.h"
#include "gla.h"
#include "interfc.h"
#include "memctrl.h"
#include "owntypes.h"
#include "random.h"
#include "sortcb.h"

/*-------------------------- DOS special -----------------------------*/

#if defined(__MSDOS__)
#include <conio.h>
#endif

/*------------------------ Type definitions --------------------------*/

typedef double*     REALVECTOR;
typedef REALVECTOR* REALCODEBOOK;

/*---------------------------  M i s c	------------------------------*/

#if defined(max) || defined(min)
#undef max
#undef min
#endif
#define  max(a,b) ((a) > (b) ? (a) : (b))
#define  min(a,b) ((a) < (b) ? (a) : (b))
#define  ScaleBetween(a,b,c) max((a), min((b), (c)))
#define  round(a)		((int) (a+0.5))

/*-------------------------- Prototypes -----------------------------*/

static double LearningRate(int,int,int);


/* =================  T I M E   F U N C T I O N S  ================= */


static long SetWatch(long* start)
{
  time_t tmp;
  return( *start = (long)time(&tmp) );
}


/* ----------------------------------------------------------------- */


static void PrintTime(long start)
{
  time_t tmp;
  long t = (long)time(&tmp) - start;
  long h,m,s;

  s = t % 60;
  t = t / 60;
  m = t % 60;
  h = t / 60;

  switch( Value(QuietLevel) )
    {
    case 0:  break;
    case 1:
    case 2:
    case 3:
    case 4:  printf("  %li:%02li:%02li\n", h, m, s);
             break;
    default: printf("Time: %li:%02li:%02li\n", h, m, s);
             break;
    }
}


/*============================  P R I N T  ===============================*/


void PrintInfo(void)
{
  printf("\n%s   %s   %s\n", ProgName, VersionNumber, LastUpdated);
  printf("Kohonen self-organizing maps (SOM) for VQ.\n");
  printf("    Usage: %s [%coption] TSfile CBfile\n", ProgName, OPTION_SYMBOL);
  printf("    For example: %s bridge tmp\n", ProgName);
  printf("\n  Options:\n");
  PrintOptions();
  printf("\n%s   %s   %s\n", ProgName, VersionNumber, LastUpdated);
}


/*-------------------------------------------------------------------*/


static void PrintOperatingInfo(char* TSName, char* CBName)
{
  if( Value(QuietLevel) >= 2 )
    {
    printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
    printf("Training Set:     %s\n", TSName);
    printf("Codebook:         %s\n", CBName);
    printf("\n");
    PrintSelectedOptions();
    printf("\n");
    }
}


/*-------------------------------------------------------------------*/


static void PrintProgress(TRAININGSET* TS,
                          CODEBOOK*    CB,
                          int          Iterations,
                          int          NeighborhoodSize,
                          long         watch)
{
  double error=0.0;

  if( Value(QuietLevel)>=2 )
     {
     error = PrintableError( AverageErrorCB(TS, CB, MSE), CB );
     }

  switch(  Value(QuietLevel) )
	{
	case 0:  return;
    case 1:  return;
    case 2:  printf("%3i: %9.4f \n", NeighborhoodSize, error);
             break;
    default: printf("%3i: %9.4f  ", NeighborhoodSize, error);
             printf("Iterated=%i  ", Iterations);
             PrintTime(watch);
             break;
	}
  fflush(stdout);
}


/*-------------------------------------------------------------------*/


static void PrintResult(TRAININGSET* TS,
                        CODEBOOK*    CB,
                        int          Iterations,
                        long         watch)
{
  double error;

  error = PrintableError( AverageErrorCB(TS, CB, MSE), CB );
  if( Value(QuietLevel)<=2 )
     {
     printf("%9.4f  (%i)", error, Iterations);
     }
  else
     {
     printf("Distortion: %-9.4f\n", error);
     printf("Iterations: %-3i\n", Iterations);
     }
  PrintTime(watch);
}


/*====================  Initialize & shutdown  ==========================*/


static void Initialize(TRAININGSET*  TS,
                       CODEBOOK*     CB,
                       PARTITIONING* P,
                       REALCODEBOOK* RCB)
{
  int i;

  if( Value(CodebookSize)>=BookSize(TS) )
      {
      printf("Codebook size (%i) > training set size (%i).\n",
              Value(CodebookSize),BookSize(TS));
      exit(-1);
      }
  CreateNewCodebook(CB, Value(CodebookSize), TS);
  AddGenerationMethod(CB, ProgName);
  AddGenerationMethod(CB, VersionNumber);
  CreateNewPartitioning(P, TS, Value(CodebookSize));
  *RCB = allocate( Value(CodebookSize) * sizeof (REALVECTOR) );
  for(i=0; i<Value(CodebookSize); i++)
     {
     (*RCB)[i] = allocate( VectorSize(TS) * sizeof(double) );
     }
  initrandom( Value(RandomSeed) );
}


/*-------------------------------------------------------------------*/


void ShutDown(TRAININGSET*  TS,
              CODEBOOK*     CB,
              PARTITIONING* P,
              REALCODEBOOK  RCB,
              int           CodebookSize)
{
  int i;

  FreeCodebook(TS);
  FreeCodebook(CB);
  FreePartitioning(P);
  for(i=0; i<CodebookSize; i++) deallocate(RCB[i]);
  deallocate(RCB);
}


/*=====================  C O D E B O O K  ==============================*/


static int FreqToIndex(TRAININGSET* TS, int sum)
{
  int i=0,l=0;

  while(l<sum) l += VectorFreq(TS, i++);
  return(i-1);
}


/*-------------------------------------------------------------------*/


static void PickRandomVector(TRAININGSET* TS, VECTORTYPE vect, int mode)
{
  long range = (TS->MaxValue - TS->MinValue);
  int  i,t,index;

  switch( mode )
     {
	 case RandomNumbers:
		{
        for(i=0; i<VectorSize(TS); i++)
            vect[i] = round( drand() * range + TS->MinValue );
        break;
        }
     case FromTSWithFreq:
        {
        t=irand(1,TotalFreq(TS));
        index=FreqToIndex(TS, t);
        CopyVector(Vector(TS,index), vect, VectorSize(TS));
		break;
        }
     case FromTSWithoutFreq:
        {
        index = irand(1,BookSize(TS)-1);
        CopyVector(Vector(TS,index), vect, VectorSize(TS));
		break;
        }
     default:
        {
        ErrorMessage("Invalid parameter: RandomBook=%i\n", Value(RandomBook));
        }
     }
}


/*-------------------------------------------------------------------*/


static void RandomCodebook(TRAININGSET* TS, REALCODEBOOK RCB)
{
  VECTORTYPE vect;
  int i,j;

  vect = CreateEmptyVector(VectorSize(TS));
  for(i=0; i<Value(CodebookSize); i++)
     {
     /* Note: does not check if duplicate! */
     PickRandomVector(TS,vect,Value(RandomBook));
     for(j=0; j<VectorSize(TS); j++)
         RCB[i][j] = vect[j];
     }
  FreeVector(vect);
}


/*-------------------------------------------------------------------*/


void CalculateDiscreteCodebook(TRAININGSET* TS,
							   REALCODEBOOK RCB,
                               CODEBOOK*    CB)
{
  int i,j;

  for(i=0; i<Value(CodebookSize); i++)
     {
     for(j=0; j<VectorSize(TS); j++)
		  {
		  VectorScalar(CB,i,j) = (VECTORELEMENT) round(RCB[i][j]);
		  }
     VectorFreq(CB,i) = 1;
     }
}


/*-------------------------------------------------------------------*/


static void OutputCodebook(char* CBName, CODEBOOK* CB)
{
  int ndups = DuplicatesInCodebook(CB);

  if( ndups>0 && Value(QuietLevel)>=2 )
     {
     printf("WARNING: Final CB contains %i duplicates.\n", ndups);
     }
  SortCodebook(CB, DATA_DESCENDING);
  WriteCodebook(CBName, CB, Value(OverWrite));
}


/*=====================  L E A R N I N G  ==============================*/


static int NearestNeuron(REALCODEBOOK  RCB,
                         VECTORTYPE    v,
                         int           VectSize,
                         int           prev)
{
  double mindist = MAXDOUBLE;
  double dist;
  int minind = 0;
  int low,high;
  int i,j;

  if( Value(LimitedSearch)==YES )
	 {
	 low   = max(0, prev-Value(InitNeighborhood));
	 high  = min(prev+Value(InitNeighborhood), Value(CodebookSize)-1);
	 }
  else
	 {
	 low   = 0;
	 high  = Value(CodebookSize)-1;
	 }

  for(i=low; i<=high; i++)
    {
    dist = 0;
    for(j=0; j<VectSize; j++)
        dist += (v[j]-RCB[i][j]) * (v[j]-RCB[i][j]);
    if(dist < mindist) { mindist = dist; minind = i; }
    }
  return(minind);
}


/*-------------------------------------------------------------------*/


void LearnVector(TRAININGSET*  TS,
                 PARTITIONING* P,
                 REALCODEBOOK  RCB,
                 int           v,
                 double        Alpha,
                 int           NeighborhoodSize)
{
  int     prev,match,low,high,i,j,d;
  double  diff;
  double  Weight=1.0;

  prev  = Map(P,v);
  match = NearestNeuron(RCB, Vector(TS,v), VectorSize(TS), prev);
  low   = max(0, match-NeighborhoodSize);
  high  = min(match+NeighborhoodSize, Value(CodebookSize)-1);
  ChangePartition(TS, P, match, v);
  Weight = Value(PartitionCount) ? 1.0/UniqueVectors(P,match) : 1.0;

  /* Increase the effect to the current neuron and its neighborhood */
  for(i=low; i<=high; i++)
	 {
     if(NeighborhoodSize>0)
		{
		d = abs(match-i);
		Weight = Weight * exp( -(double)d / NeighborhoodSize );
		}
     for(j=0; j<VectorSize(TS); j++)
		{
        diff = (double) VectorScalar(TS,v,j) - RCB[i][j];
        RCB[i][j] += VectorFreq(TS,v) * Alpha * Weight * diff;
		}
	 }

  /* Reduce the effect from the previous neuron */
  if( Value(DecreasePrevious) && prev!=match )
     {
     for(j=0; j<VectorSize(TS); j++)
		{
        diff = (double) VectorScalar(TS,v,j) - RCB[prev][j];
        RCB[prev][j] -= VectorFreq(TS,v) * Alpha * diff;
		}
     }
}        


/*-------------------------------------------------------------------*/


static int NumberOfIterations(int NeighborhoodSize)
{
  int iter;

  if( Value(AdaptiveIterations) )
       iter = max( 1, Value(TSIterations) >> NeighborhoodSize );
  else iter = Value(TSIterations);
/* printf("ns=%i, tsiter=%i  iter=%i \n",NeighborhoodSize, Value(TSIterations), iter); */
  return(iter);
}


/*-------------------------------------------------------------------*/


static double LearningRate(int SoFar, int SoFarTotal, int ThisTimeTotal)
{
  double maxrate = (double) Value(LearningRateMax) / Max(LearningRateMax);
  int	 iter	 = Value(InitNeighborhood);
  int    total;
  double ratio;

  if( Value(LearningCurve)==ExpReset ) total = ThisTimeTotal;
  else if( Value(AdaptiveIterations) ) total = 2 * Value(TSIterations);
  else                                 total = (iter+2) * Value(TSIterations);
  if( SoFarTotal>=total )              total = SoFarTotal-1;

  switch( Value(LearningCurve) )
     {
     case Linear:   ratio = (double) (total-SoFarTotal) / total;  break;
     case Exp:      ratio = exp( -(double)SoFarTotal / total);    break;
     case ExpReset: ratio = exp( -(double)SoFar / total);         break;
     default:       ratio = (double) (total-SoFarTotal) / total;  break;
     }
  ratio = maxrate * ratio;
  if( Value(QuietLevel)>=5 )
     printf("SoFar=%i, SoFarTotal=%i ThisTimeTotal=%i  Total=%i  max=%5.2f  alpha=%5.2f \n",
     SoFar, SoFarTotal, ThisTimeTotal, total, maxrate, ratio);
  return(ratio);
}


/*-------------------------------------------------------------------*/


static void LearnTrainingSet(TRAININGSET*  TS,
							 CODEBOOK*	   CB,
                             PARTITIONING* P,
							 REALCODEBOOK  RCB,
                             int           NeighborhoodSize,
                             int*          Iterations,
                             long          watch)
{
  double  Alpha;
  int	  i,j;

  for(i=0; i<NumberOfIterations(NeighborhoodSize); i++)
     {
     Alpha = LearningRate(i, *Iterations, NumberOfIterations(NeighborhoodSize));
     for(j=0; j<BookSize(TS); j++)
        {
        LearnVector(TS, P, RCB, j, Alpha, NeighborhoodSize);
        }
     (*Iterations)++;
     if( Value(QuietLevel)>=3 )
        {
        CalculateDiscreteCodebook(TS, RCB, CB);
        PrintProgress(TS, CB, *Iterations, NeighborhoodSize, watch);
        }
     }

  if( Value(QuietLevel)<=2 )
	 {
	 CalculateDiscreteCodebook(TS, RCB, CB);
	 PrintProgress(TS, CB, *Iterations, NeighborhoodSize, watch);
	 }
}


/*===========================  M A I N  ==============================*/


int main(int argc, char* argv[])
{
  TRAININGSET   TS;
  CODEBOOK		CB;
  PARTITIONING  P;
  REALCODEBOOK	RCB;
  char          TSName[64] = { '\0' };
  char          CBName[64] = { '\0' };
  long          watch;
  int           GoOn=YES;
  int           NeighborhoodSize;
  int			Iterations=0;
  int			GLAIter;
  double        Error1, Error2;

  ParameterInfo paraminfo[3] = { { TSName, FormatNameTS, 0, INFILE  },
                                 { CBName, FormatNameCB, 0, OUTFILE } };

  ParseParameters(argc, argv, 2, paraminfo);
  initrandom(Value(RandomSeed));
  PrintOperatingInfo(TSName, CBName);
  ReadTrainingSet(TSName, &TS);
  Initialize(&TS, &CB, &P, &RCB);
  NeighborhoodSize = Value(InitNeighborhood);
  SetWatch(&watch);

  RandomCodebook(&TS, RCB);
  CalculateDiscreteCodebook(&TS, RCB, &CB);
  PrintProgress(&TS, &CB, 0, 0, watch);

  while( NeighborhoodSize>=0 && GoOn )
	 {
     LearnTrainingSet(&TS, &CB, &P, RCB, NeighborhoodSize, &Iterations, watch);
     NeighborhoodSize--;
	 #if defined(__MSDOS__)
	 if (kbhit()) GoOn=(getch()!=27);
	 #endif
	 }

  CalculateDiscreteCodebook(&TS, RCB, &CB);
  if( Value(GLAIterations) )
     {
	 GLA(&TS, &CB, NULL, &P, Value(GLAIterations), &Error1, &Error2, &GLAIter);
     }
  PrintResult(&TS, &CB, Iterations, watch);
  OutputCodebook(CBName, &CB);
  ShutDown(&TS, &CB, &P, RCB, Value(CodebookSize));
  checkmemory();
  return( 0 );
}

